;(function(root) {
  var tincan;
  var statementQueue = [];

  var state = {
    completed: false,
    prevDuration: null,
    score: null,
    startTime: null
  }

  var BOOKMARK = 'bookmark';
  var SUSPEND_DATA = 'suspend_data';
  var TOTAL_TIME = 'cumulative_time';

  var getUUID = root.TinCan.Utils.getUUID;
  var formatTime = root.TinCan.Utils.convertMillisecondsToISO8601Duration;
  var noop = Function.prototype;

  var quizzes   = null;
  var quizcount = 0;
  var passcount = 0;
  var oomcount  = 0;
  var unitprog  = 0;
  var unitscore = 0;

  var passsent  = false;
  var oomsent   = false;
  var scoresent = 0;
  var progsent  = 0;

  var requestvars = window.parent.location.search.replace('?', '').split('&');
  var forcesync = requestvars.includes('resend');

  function syncQuizzes(qdata) {

    if(quizzes == null) {
      var dcourse = JSON.parse(root.Base64.fromBase64(root.window.courseData));
      var dlessons = dcourse.course.lessons;

      quizzes = new Object();
      for (i = 0; i < dlessons.length; ++i) {
        if(dlessons[i].type == 'quiz') {
          quizcount ++;
          quizzes[i] = {
            'id' : dlessons[i].id,
            'title' : dlessons[i].title,
            'passed' : false,
            'score' : 0,
          };
        }
      }
    }

    var lessons  = qdata.progress.lessons;
    var sumscore = 0;

    passcount = 0;
    oomcount  = 0;
    unitscore = 0;

    for(var key in quizzes) {
      if (lessons.hasOwnProperty(key)) {
        var lesson = lessons[key];
        if (lesson.hasOwnProperty('pq')) {
          quizzes[key].passed = lessons[key].pq;
        }
        if (lesson.hasOwnProperty('s')) {
          if(quizzes[key].score < lessons[key].s) {
            quizzes[key].score = lessons[key].s;
          }
        }
      }

      if(quizzes[key].passed) {
        passcount++;
        sumscore += quizzes[key].score;

        if(quizzes[key].score >= 100) oomcount++;
      }
    }

    unitprog  = passcount > 0 ? (passcount / quizcount) * 100 : 0;
    unitscore = passcount > 0 ? sumscore / passcount : 0;
  }

  function isFile() {
    return /file/.test(root.location.protocol)
  }

  function assign(target) {
    if (target === undefined || target === null) {
      throw new TypeError('assign: Cannot convert undefined or null to object');
    }

    var output = Object(target);

    var source;
    var nextKey;

    for (var index = 1; index < arguments.length; index++) {
      source = arguments[index];

      if (source !== undefined && source !== null) {
        for (nextKey in source) {
          if (source.hasOwnProperty(nextKey)) {
            output[nextKey] = source[nextKey];
          }
        }
      }
    }

    return output;
  }

  function debounce(fn, delay) {
    var timer = null;

    return function () {
      var context = this, args = arguments;
      clearTimeout(timer);
      timer = setTimeout(function () {
        fn.apply(context, args);
      }, delay);
    };
  }

  function closeContent() {
    if(root.top === root) {
      root.close();
    }
    else {
      root.location.pathname =
        root.location.pathname
          .split('/')
          .slice(0, -1)
          .join('/')
          .concat('/goodbye.html');
    }
  }

  function queueStatement(stmt) {
    if(statementQueue.length) {
      statementQueue.unshift(stmt);
    } else {
      sendStatement(stmt);
    }
  }

  function popStatement(stmt) {
    if(statementQueue.length) {
      sendStatement(statementQueue.pop());
    }
  }

  function sendStatement(stmt) {
    tincan.sendStatement(wrapStatement(stmt), popStatement);
  }

  function wrapStatement(stmt) {
    return assign({}, {
      id: getUUID(),
      type: 'http://adlnet.gov/expapi/activities/course',
      object: { id: tincan.activity.id }
    }, stmt)
  }

  function loadBookmark() {
    var hash = GetBookmark();

    if(!hash) {
      return;
    }

    if(root.history.pushState) {
      root.history.pushState(null, null, hash);
    } else {
      root.location.hash = hash;
    }
  }

  function sendAttempted() {
    commitData({
      verb: 'attempted'
    });
  }

  function sendExperienced(completion) {
    commitData({
      verb: 'experienced',
      result: { completion: completion }
    });
  }

  function addFinishData(duration, stmt) {
    const result = assign({}, stmt.result);

    result.duration = formatTime(duration);

    if(state.score !== null) {
      result.score = state.score;
    }

    return assign({}, stmt, { result: result });
  }

  function getDuration() {
    if(state.prevDuration === null) {
      state.prevDuration = Number(getState(TOTAL_TIME)) || 0
    }

    state.prevDuration = state.prevDuration + ((new Date()).getTime() - state.startTime)

    state.startTime = (new Date()).getTime();

    return state.prevDuration;
  }

  function commitData(stmt) {
    var duration = getDuration();

    queueStatement(addFinishData(duration, stmt));

    // run last as we need to have a sync call for beforeunload
    // called with no cfg.callback
    setState(TOTAL_TIME, duration);
  }

  function getState(key, cfg) {
    try {
      var state =
        tincan.getState(key, cfg || {}).state;

      return state && state.contents
        ? state.contents
        : '';
    }
    catch(ex) {
      return '';
    }
  }

  function setState(key, data, cfg) {
    try {
      tincan.setState(key, data, cfg || {});
    }
    catch(ex) {
      return;
    }
  }

  function SetBookmark(data) {
    setState(BOOKMARK, data, { callback: noop });
  }

  function GetBookmark() {
    return getState(BOOKMARK);
  }

  function GetTime() {
    return getState(TOTAL_TIME);
  }

  function SetTime() {
    return setState(TOTAL_TIME);
  }

  function compress(data) {
    try {
      return JSON.stringify({
        v: 1,
        d: lzwCompress.pack(JSON.stringify(data))
      });
    } catch(e) {
      return '';
    }
  }

  function decompress(dstring) {
    try {
      const data = JSON.parse(dstring);

      return data.d
        ? JSON.parse(lzwCompress.unpack(data.d))
        : data;
    } catch(e) {
      return {};
    }
  }

  function decodeHtml(html) {
      var txt = document.createElement("textarea");
      txt.innerHTML = html;
      return txt.value;
  }

  function IPMRestCall(method) {

    var payload = new Object();
    payload.activity  = tincan.activity.id;
    payload.title     = tincan.activity.definition.name['en-US'];
    payload.actor     = new Object();
    if(tincan.actor.mbox) {
      payload.actor.mode  = 'external';
      payload.actor.name  = decodeHtml(tincan.actor.name);
      payload.actor.email = tincan.actor.mbox.substring(7);
    } else if(tincan.actor.account) {
      payload.actor.mode  = 'internal';
      payload.actor.name  = decodeHtml(tincan.actor.name);
      payload.actor.uid   = tincan.actor.account.name;
    } else {
      return;
    }

    payload.progress = unitprog;
    payload.score    = unitscore;
    payload.passed   = passsent;
    payload.oomed    = oomsent;
    payload.qdata    = quizzes;

    var xmlhttp = new XMLHttpRequest();

    xmlhttp.onreadystatechange = function() {
        if (xmlhttp.readyState == XMLHttpRequest.DONE) {
           if (xmlhttp.status == 200) {
               console.log('OPENLink: '+xmlhttp.response);
           }
           else if (xmlhttp.status == 400) {
               console.log('IPMRestCall: error 400');
           }
           else {
               console.log('IPMRestCall: error');
           }
        }
    };

    xmlhttp.open("POST", "https://open.institute.pm/wp-json/ipm/tincan/"+method, true);
    xmlhttp.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
    xmlhttp.send('payload='+JSON.stringify(payload));
  }

  function SetDataChunk(data) {
    var qdata = data ? decompress(data) : {};
    var method = 'update';

    if(passsent)  qdata.passsent  = true;
    if(oomsent)   qdata.oomsent   = true;
    if(scoresent) qdata.scoresent = scoresent;
    if(progsent)  qdata.progsent  = progsent;
    if(unitprog)  qdata.unitprog  = unitprog;
    if(unitscore) qdata.unitscore = unitscore;

    //console.log('Host: '+window.location.hostname);

    if(forcesync || !qdata.passsent || !qdata.oomsent || !qdata.scoresent || qdata.scoresent < unitscore || qdata.progsent < unitprog) {
      syncQuizzes(qdata);

      if(quizcount > 0 && passcount >= quizcount) {
        var dosend = false;

        if(!qdata.passsent) {
          passsent = true;
          qdata.passsent = true;
          dosend = true;
          method='passed';
        }
        if(oomcount >= quizcount && !qdata.oomsent) {
          oomsent = true;
          qdata.oomsent = true;
          passsent = true;
          qdata.passsent = true;
          dosend = true;
          method='oomed';
        }

        if(dosend) {
          SetPassedReal();
        }
      }

      if(passcount > 0 && (forcesync || !qdata.scoresent || unitscore > qdata.scoresent || !qdata.progsent || unitprog > qdata.progsent)) {
        qdata.scoresent = unitscore;
        scoresent = unitscore;
        qdata.progsent = unitprog;
        progsent = unitprog;
        IPMRestCall(method);
        forcesync = false;
      }
    }

    setState(SUSPEND_DATA, compress(qdata), { callback: noop });
  }

  function GetDataChunk() {
    return getState(SUSPEND_DATA);
  }

  function SetScore(newScore, max, min) {
    state.score = {
      raw: newScore,
      max: max,
      min: min
    };
  }

  function InitCompletion(isComplete) {
    state.completed = isComplete;
  }

  function SetFailed() {
    state.completed = true;

    commitData({
      verb: 'failed',
      result: {
        completion: state.completed,
        success: false
      }
    });
  }

  function SetPassedReal() {
    state.completed = true;

    commitData({
      verb: 'passed',
      result: {
        completion: state.completed,
        success: true,
        extensions: {
          'https://open.institute.pm/state' : quizzes
        }
      }
    });
  }

  function SetPassed() {
    /*
    state.completed = true;

    commitData({
      verb: 'passed',
      result: {
        completion: state.completed,
        success: true
      }
    });
    */
  }

  // Record Answer Data
  function buildDescriptionObj(id, title) {
    return {
      id: id,
      description: {
        und: title
      }
    }
  }

  function prop(property) {
    return function (obj) {
      return obj[property];
    }
  }

  function mapMatchingArray(match) {
    return match.source.id + '[.]' + match.target.id;
  }

  function joinArray(a) {
    return a.join('[,]');
  }

  function getSources(answers) {
    var mapAnswers = function(answer) {
      return buildDescriptionObj(answer.source.id, answer.source.title);
    };

    return answers.map(mapAnswers);
  }

  function getTargets(answers) {
    var mapAnswers = function(answer) {
      return buildDescriptionObj(answer.target.id, answer.target.title);
    };

    return answers.map(mapAnswers);
  }

  function getChoices(answers) {
    var mapAnswers = function(answer) {
      return buildDescriptionObj(answer.id, answer.title);
    };

    return answers.map(mapAnswers);
  }

  function getCorrectResponsesPattern(data) {
    var correctResponse = data.correctResponse;

    switch (data.type) {
      case 'FILL_IN_THE_BLANK':
        return correctResponse;
      case 'MATCHING':
        return [ joinArray(correctResponse.map(mapMatchingArray)) ];
      case 'MULTIPLE_CHOICE':
      case 'MULTIPLE_RESPONSE':
      default:
        return [ joinArray(correctResponse.map(prop('id'))) ];
    }
  }

  function buildDefinition(data) {
    var answers = data.answers;
    var correctResponse = data.correctResponse;
    var title = { und: data.questionTitle };
    var type = data.type;

    var interactionTypes = {
      FILL_IN_THE_BLANK: 'fill-in',
      MATCHING: 'matching',
      MULTIPLE_CHOICE: 'choice',
      MULTIPLE_RESPONSE: 'choice'
    };

    var definition = {
      name: title,
      description: title,
      type: 'http://adlnet.gov/expapi/activities/cmi.interaction',
      interactionType: interactionTypes[type || 'MULTIPLE_CHOICE'],
      correctResponsesPattern: getCorrectResponsesPattern(data)
    };

    if (type === 'MULTIPLE_CHOICE' || type === 'MULTIPLE_RESPONSE') {
      definition.choices = getChoices(answers);
    }

    if (type === 'MATCHING') {
      definition.source = getSources(answers);
      definition.target = getTargets(answers);
    }

    return definition;
  }

  function buildResponse(data) {
    var response = data.response;

    switch (data.type) {
      case 'FILL_IN_THE_BLANK':
        return response;
      case 'MATCHING':
        return joinArray(response.map(mapMatchingArray));
      case 'MULTIPLE_CHOICE':
      case 'MULTIPLE_RESPONSE':
      default:
        return joinArray(response.map(prop('id')));
    }
  }

  function ReportAnswer(data) {
    var definition = buildDefinition(data);
    var response = buildResponse(data);
    var title = data.questionTitle;

    commitData({
      type: 'http://adlnet.gov/expapi/activities/cmi.interaction',
      verb: 'answered',
      description: title,
      name: title,
      object: {
        id: tincan.activity.id + '/' + data.itemId + '_' + Date.now(),
        definition: definition
      },
      result: {
        success: data.isCorrect,
        response: response
      }
    });
  }

  function ConcedeControl() {
    sendExperienced(state.completed);
    root.removeEventListener('beforeunload', ConcedeControl);
    closeContent();
  }

  function lms() {
    root.removeEventListener('beforeunload', ConcedeControl);
    root.addEventListener('beforeunload', ConcedeControl);

    var url =
      root.location.href.replace(root.location.hash, '');

    var config = {
      activity: {
        id: TC_COURSE_ID,
        definition: {
          name: TC_COURSE_NAME,
          description: TC_COURSE_DESC
        }
      }
    }

    tincan = new TinCan(isFile()
      ? config
      : assign({}, config, { url: url })
    );

    state.startTime = (new Date()).getTime();

    loadBookmark();
    sendAttempted();

    return {
      SetBookmark: SetBookmark,
      SetDataChunk: SetDataChunk,
      GetDataChunk: GetDataChunk,
      InitCompletion: InitCompletion,
      ReportAnswer: ReportAnswer,
      SetFailed: SetFailed,
      SetPassed: SetPassed,
      SetScore: SetScore,
      ConcedeControl: ConcedeControl,
      utils: {
        assign: assign,
        debounce: debounce
      }
    };
  }

  root.lms = lms;

}(window));